"""This is a library that works with mod_python to allow you to use
OpenID as access control for any resource that Apache serves. See
README for more details."""

__copyright__ = """Copyright (C) 2008 Canonical, Inc.
Copyright (C) 2005, 2006 JanRain, Inc."""

__license__ = """This is Free Software, under the terms of the GNU GPL.
See COPYING for details."""

__version__ = '1.99.0'

def protect(apache_request):
    """This is the function that protects a directory.

    It should be used as a PythonAccessHandler. e.g.:

      PythonAccessHandler mpopenid::protect

    Options used by this handler:

      # whitespace separated OpenID identifiers
      PythonOption authorized-users ""

      # a URL to a list of OpenID identifiers
      PythonOption authorized-users-list-url ""

    If both authorized-users and authorized-users-list-url are specified,
    the lists will be concatenated. If neither is specified, no users
    will have access. (this is not very useful)
    """
    oid_req = OpenIDProtect(apache_request)
    try:
        return oid_req.protect()
    finally:
        oid_req.session.save()

def openid(apache_request):
    """This is the function that handles OpenID logins

    It should be used as a PythonHandler. e.g.:

      SetHandler mod_python
      PythonHandler mpopenid::openid

    Options used by this handler:

      # Defaults to "memory"
      # options are file, memory, stateless
      # stateless only works with python-openid 1.x
      PythonOption store-type file

      # Used only if it's a file store
      PythonOption store-directory /somewhere/

      # URL to redirect to after logout
      PythonOption external-logout-url http://someurl/

      # List of external cookies that should be deleted on logout
      # A path can be encoded in a cookie's name using ':'
      PythonOption external-cookie-names acookie bcookie /some/path:ccookie
    """
    oid_req = OpenIDLogin(apache_request)
    try:
        return oid_req.openid()
    finally:
        oid_req.session.save()

##### The rest of this file is support for the functions above #####


import os
import re
import sys
import time
import socket
import urllib
import datetime
import urlparse
import xmlrpclib

from cgi import escape
from string import Template

try:
    from mod_python import apache, util, Cookie, Session
except ImportError:
    print >>sys.stderr, ("Unable to import mod_python code. Continuing "
                         "anyway (for pychecker)")
    sys.stderr.flush()
    apache = util = Cookie = Session = None

try:
    from openid.consumer import consumer
    from openid.association import Association
except ImportError:
    print >>sys.stderr, (
        'No JanRain OpenID library was found. Checked in:\n' + str(sys.path))
    raise

try:
    from openid.urinorm import urinorm
except ImportError:
    # An OpenID library that does not yet have urinorm. Anything
    # before 2.0 should be in this state. Not normalizing URLs just
    # means that you have to be more careful about what you type when
    # you're listing the authorized URLs.
    from openid.oidutil import normalizeUrl as urinorm

try:
    # python-openid 1.2, 2.x
    from openid.yadis.discover import DiscoveryFailure
    from openid.fetchers import HTTPFetchingError
    from openid.store import filestore, memstore
except ImportError:
    # python-openid 1.1
    from yadis.discover import DiscoveryFailure
    from urljr.fetchers import HTTPFetchingError
    from openid.store import filestore, dumbstore

# a version of the openid library that supports Launchpad teams
try:
    from openid.extensions import teams
    teams_extension_avail = True
except ImportError:
    teams_extension_avail = False

# BarryWarsaw 18-Dec-2008 FOR TESTING ONLY DO NOT DEPLOY
#
# When testing this stuff locally, enable the following.  This uses a url
# fetcher that isn't as picky about SSL problems.  You should generally want
# the standard behavior here.
#
#from openid.fetchers import setDefaultFetcher, Urllib2Fetcher
#setDefaultFetcher(Urllib2Fetcher())

expected_discovery_exceptions = (DiscoveryFailure, HTTPFetchingError)

message_text = {
    'empty':('message', 'Enter an OpenID URL to continue'),
    'cancel':('message', 'Authorization cancelled'),
    'bad_html':('error', 'No OpenID information was found at that URL'),
    'http_failed':('error',
                   'There was an error communicating with the server'),
    'failure':('error', 'The server reported an error'),
    'denied':('error', 'That identity URL has not been granted '
                       'access to this resource'),
    'discovery':('error', 'Failed to discover an OpenID server'),
    }

login_page_tmpl = '''\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
          "http://www.w3.org/TR/html4/strict.dtd">
<html>
  <head>
    <title>%(title)s</title>
    <style type="text/css">
      div.error {
          background: #ffdddd;
          border: 1px solid red;
          padding: 0.5em;
      }
      div.message {
          background: #ffffdd;
          border: 1px solid yellow;
          padding: 0.5em;
      }
      form {
          margin-top: 1em;
      }
    </style>
  </head>
  <body onload="%(onload)s">
    <h1>%(title)s</h1>
    %(resource)s%(message)s
    <form method="post" action="%(action)s" name="oid_form">
      <div>
        %(form_input)s
        <input id="continue_btn" type="submit" value="Continue">
      </div>
    </form>
  </body>
</html>
'''

post_redirect_tmpl = '''\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
          "http://www.w3.org/TR/html4/strict.dtd">
<html>
  <head>
    <title>%(title)s</title>
    <script type="text/javascript">
      function autoSubmit() {
          var inputs = document.getElementsByTagName("input");
          for (var i = 0; i < inputs.length; i++) {
              inputs[i].style.visibility = "hidden";
          }

          document.getElementById("%(id)s").submit();
      }
    </script>
  </head>
  <body onload="autoSubmit()">
    %(form)s
  </body>
</html>
'''

def parse_authorized_urls(data):
    """Parse and normalize a string containing a whitespace-separated
    list of URLs"""
    urls = []
    for s in data.split():
        url = s.strip()
        if not url or url[0] == '#':
            continue
        parsed = urlparse.urlparse(url)
        if not (parsed[0] and parsed[1]):
            url = 'http://' + url
        urls.append(urinorm(url))
    return urls

def split_values(data):
    """Parse a string containing whitespace-separated values
    """
    values = []
    for s in data.split():
        val = s.strip()
        if not val or val[0] == '#':
            continue
        values.append(val)
    return values

def parse_allowed_ops(data):
    """Parse a key=value string of OP URI's, separated by newlines
    """
    ops = {}
    for s in data.split('\n'):
        op = s.strip()
        if not op or op[0] == '#':
            continue
        op_parts = op.split('=')
        ops[op_parts[0].strip()] = op_parts[1].strip()
    return ops

class OpenIDAccessRequest(object):
    # Cache for store instances
    _file_stores = {}
    _max_store_cache_size = 5

    # Cache for authorized identities
    _authorized_cache = {}
    _max_authorized_cache_size = 100

    # Cache for authorized teams
    _teams_cache = {}
    _max_teams_cache_size = 100

    # Time out sessions after a week of no access
    session_timeout = 60 * 60 * 24 * 7

    def __init__(self, apache_request):
        self.apache_request = apache_request
        self.options = apache_request.get_options()
        self.session = Session.Session(
            apache_request,
            timeout=self.session_timeout,
            lock=False)

    def __getattr__(self, attr):
        func = object.__getattribute__(self, 'get_' + attr)
        val = func()
        setattr(self, attr, val)
        return val

    def debug(self, message, *args):
        """Write debugging output to the Apache error log.

        Output only occurs if this line appears in the configuration file:

        PythonOption debug 1

        :param message: The %-style string template.
        :type message: string
        :param args: An optional sequence of values to interpolate into
            the `message` template.
        :type args: A *-args tuple
        """
        if self.options.get('debug', '0') == '1':
            self.error(message, *args)

    def error(self, message, *args):
        """Write output to the Apache error log.

        :param message: The %-style string template.
        :type message: string
        :param args: An optional sequence of values to interpolate into
            the `message` template.
        :type args: A *-args tuple
        """
        self.apache_request.log_error(message % args)

    @property
    def cache_lifetime(self):
        """The team cache lifetime from the config options."""
        cache_lifetime = self.options.get('team-cache-lifetime', 3600)
        try:
            cache_lifetime = int(cache_lifetime)
        except ValueError:
            self.error('Bad team-cache-lifetime: %s', cache_lifetime)
            cache_lifetime = 3600
        self.debug('team-cache-lifetime: %s', cache_lifetime)
        return datetime.timedelta(seconds=cache_lifetime)

    def get_cookied_user(self):
        """Get the openid_identifier cookie for this request

        mod_python.Request -> NoneType or str
        """
        return self.session.get('cookied_user')

    def set_cookied_user(self, openid_identifier):
        """Set the openid_identifier cookie for this user

        (mod_python.Request, str, int) -> NoneType
        """
        self.session['cookied_user'] = openid_identifier

    def del_cookied_user(self):
        """Delete the user session."""
        del self.session['cookied_user']

    cookied_user = property(get_cookied_user, set_cookied_user,
                            del_cookied_user, get_cookied_user.__doc__)

    def get_cookied_teams(self):
        """Get the list of Launchpad teams for the current user.

        The cookied teams list is cached for a period of time, specified by
        the team-cache-lifetime option.
        """
        # BarryWarsaw 18-Dec-2008 At one point I tried to use a separate class
        # to represent the cookied team cache, but that had several problems.
        # Because the session data is actually pickled into the dbm file, the
        # instance stored here must be picklable.  However, in order to do
        # logging in the cache instance, you'd have to pass a hook into that
        # instance which would not be picklable (e.g. an instancemethod or the
        # Apache request object).  So then you'd have to deal with the pickle
        # protocol and figure out a way to initialize that log hook when the
        # session is unpickled.  Or forgo logging.  Neither option is really
        # acceptable, so the choice was made to use a simpler inline cache.
        # However this means that we need to hijack assignment to
        # .cookied_teams in order to keep the cache consistent.
        team_cache = self.session.setdefault('cookied_teams', {})
        # Evict expired cache items.  Use .keys() to avoid
        # mutation-while-iterating exceptions.
        now = datetime.datetime.now()
        for team in team_cache.keys():
            if team_cache[team] < now:
                self.debug('Evicting cookied team: %s', team)
                del team_cache[team]
        return team_cache.keys()

    def set_cookied_teams(self, teams):
        """Set the list of Launchpad teams for the current user.

        We pervert the meaning of assignment here to be extension.  Use this
        instead of .extend() on self.cookied_teams because the latter will not
        be persistent.  Practicality beats purity.
        """
        expiration_date = datetime.datetime.now() + self.cache_lifetime
        team_cache = self.session.setdefault('cookied_teams', {})
        team_cache.update(dict((team, expiration_date) for team in teams))

    def del_cookied_teams(self):
        """Delete the list of Launchpad teams for the current user."""
        self.session.setdefault('cookied_teams', {}).clear()

    cookied_teams = property(get_cookied_teams, set_cookied_teams,
                             del_cookied_teams, get_cookied_teams.__doc__)

    def get_server_url(self):
        """Return a URL to the root of the server that is serving this
        request.

        mod_python.Request -> str
        """
        host = self.apache_request.hostname
        port = self.apache_request.connection.local_addr[1]

        # Don't include the default port number in the URL
        try:
            # mod_python 3.3
            https = self.apache_request.is_https()
        except:
            # mod_python 3.1
            # subprocess_env values are not reliable, so it is
            # not likely that you can use a custom https port
            # unless you upgrade to mod_python 3.3
            # http://www.modpython.org/pipermail/mod_python/2006-February/020197.html
            self.apache_request.add_common_vars()
            if port == 443 or self.apache_request.subprocess_env.get('HTTPS', 'off') == 'on':
                https = True
            else:
                https = False

        if https:
            default_port = 443
            proto = 'https'
        else:
            default_port = 80
            proto = 'http'

        if port == default_port:
            server_url = '%s://%s/' % (proto, host)
        else:
            server_url = '%s://%s:%s/' % (proto, host, port)

        return server_url

    def get_request_uri(self):
        # apache_request.request_uri and unparsed_uri
        # this may not always be giving the correct value!
        # Seems to miss the "/openid/" prefix when called
        # inside a <Location /openid/> directive.
        # The value in subprocess_env seems correct.
        #return urlparse.urljoin(self.server_url, self.apache_request.unparsed_uri)
        return urlparse.urljoin(self.server_url,
            self.apache_request.subprocess_env['REQUEST_URI'])

    def get_action_path(self):
        raise NotImplementedError

    @property
    def _substitutions(self):
        """A dictionary of substitutions.

        $-substitutions are allowed, using the CGI environment, plus the
        public attributes of the request object, plus session data, in that
        order of preference.
        """
        substitutions = self.apache_request.subprocess_env.copy()
        for attribute in dir(self.apache_request):
            if attribute.startswith('_'):
                continue
            value = getattr(self.apache_request, attribute)
            if isinstance(value, basestring):
                substitutions[attribute] = value
        substitutions.update(self.session)
        self.debug('substitutions: %r', substitutions)
        return substitutions

    def scheme_default(self, authorized_list_url):
        """Fetch a list of teams using GET.

        This uses `urllib.urlopen()` so any scheme or url supported by that
        method is supported here.  Typically this will be `http`.  GET must
        return a file containing a whitespace-separated list of team names.

        :param authorized_list_url: The url to fetch from.
        :type authorized_list_url: string
        :return: A list of team names.
        :rtype: list of strings
        """
        try:
            url_handle = urllib.urlopen(authorized_list_url)
        except IOError:
            self.error('Failed to fetch authorized teams URL %r',
                       authorized_list_url)
            sys.excepthook(*sys.exc_info())
            sys.stderr.flush()
        else:
            data = url_handle.read()
            url_handle.close()
            return split_values(data)

    def scheme_regexp(self, authorized_list_url):
        """Fetch a list of teams using regexp matching.

        This method performs a regular expression match (using Python's regexp
        syntax) against a string defined in the regexp-target option.  That
        option is a $-style string template which gets substitutions applied
        from the CGI environment, plus the public attributes of the request
        object, plus session data, in that order of preference.

        The regexp-group option names a regexp group that will return a team
        name when a successful match occurs.

        :param authorized_list_url: A url of the form regexp:<pattern> where
            <pattern> is any valid Python regular expression.
        :type authorized_list_url: string
        :return: A list of team names.
        :rtype: list of strings
        """
        # len('regexp:') == 7
        pattern = authorized_list_url[7:]
        self.debug('pattern: %r', pattern)
        target = self.options.get('regexp-target')
        if not target:
            self.error('No regexp target')
            return []
        self.debug('target (raw): %r', target)
        target = Template(target).safe_substitute(self._substitutions)
        self.debug('target (cooked): %r', target)
        # Now do regular expression match to get the team name.
        try:
            match_object = re.search(pattern, target)
        except:
            self.error('Bad regular expression pattern: %r [target: %r]',
                       pattern, target)
            sys.excepthook(*sys.exc_info())
            sys.stderr.flush()
        if match_object is None:
            self.error('Pattern failed to match, ignoring.')
            return []
        # Extract the group.
        group = self.options.get('regexp-group')
        if not group:
            self.debug('No regexp group found, returning full match')
            group = 0
        self.debug('Using group: %r', group)
        team = match_object.group(group)
        if not team:
            self.error('No team group value found')
            return []
        else:
            self.debug('Group team: %r', team)
            return [team]

    def scheme_xmlrpc(self, authorized_list_url):
        """Fetch a list of teams using XMLRPC.

        This method performs an XMLRPC call, which must return a list of team
        names.  The XMLRPC end-point is described in the option, with a
        leading 'xmlrpc:' scheme identifier, e.g.

        PythonOption authorized-teams-list-url xmlrpc:http://api.example.com/endpoint

        Everything after the 'xmlrpc:' is used as the end-point.  The method
        to call is named in the `xmlrpc-method` option, and arguments to pass
        to the method are named in the `xmlrpc-argument` option.  This latter
        is a $-style string template which gets substitutions applied from the
        CGI environment, plus the public attributes of the request object,
        plus session data, in that order of preference.

        If `xmlrpc-arguments` is optional, in which case the method is called
        with no arguments.

        :param authorized_list_url: A url of the form xmlrpc:<endpoint> where
            <endpoint> is any valid XMLRPC end point.
        :type authorized_list_url: string
        :return: A list of team names.
        :rtype: list of strings
        """
        # len('xmlrpc:') == 7
        end_point = authorized_list_url[7:]
        self.debug('XMLRPC end point: %r', end_point)
        method = self.options.get('xmlrpc-method')
        if not method:
            self.error('No XMLRPC method')
            return []
        self.debug('xmlrpc method: %s', method)
        missing = object()
        argument = self.options.get('xmlrpc-argument', missing)
        if argument is not missing:
            self.debug('argument (raw): %r', argument)
            argument = Template(argument).safe_substitute(self._substitutions)
            self.debug('argument (cooked): %r', argument)
        # Now make the XMLRPC call.
        try:
            server = xmlrpclib.ServerProxy(end_point)
            if argument is missing:
                return getattr(server, method)()
            else:
                return getattr(server, method)(argument)
        except:
            self.error('XMLRPC method call failed')
            sys.excepthook(*sys.exc_info())
            sys.stderr.flush()
            return []

    def get_authorized_teams(self):
        """Get all authorized teams for this request."""
        authorized_s = self.options.get('authorized-teams')
        if authorized_s is None:
            self.debug('No authorized-teams')
            authorized_list = []
        else:
            try:
                authorized_list = self._teams_cache[authorized_s]
            except KeyError:
                if len(self._teams_cache) == self._max_teams_cache_size:
                    self._teams_cache.clear();

                authorized_list = split_values(authorized_s)
                self._teams_cache[authorized_s] = authorized_list

        authorized_list_url = self.options.get('authorized-teams-list-url')
        if authorized_list_url is not None:
            self.debug('authorized-teams-list-url: %r', authorized_list_url)
            parts = urlparse.urlparse(authorized_list_url)
            # Use the url scheme to determine how to get the team list.  We
            # have a few custom ones (e.g. regexp: and xmlrpc:) and Python
            # itself provides most defaults through urllib.
            method = getattr(self, 'scheme_' + parts[0], self.scheme_default)
            teams = method(authorized_list_url)
            authorized_list.extend(teams)
        return authorized_list

    def check_public_teams(self):
        """Check that all authorized teams are also public teams.

        Uses the session to cache the list of public teams, reducing the load
        on the XMLRPC server.  The cache expires after a set amount of time.

        :return: True if all the authorized teams are public.
        :rtype: boolean
        """
        # BarryWarsaw 12-Dec-2008 this could be rather inefficient, but
        # currently Launchpad's XMLRPC only allows us to check one team at a
        # time, and for our primary use case, we only need to check one team,
        # so practically speaking right now it makes no difference.
        cache = self.session.setdefault('public_team_cache', {})
        # The cache maps public team names to the date we should evict the
        # team name from the cache.  Start by seeing if any of the authorized
        # teams are in the cache, and if the cache is still valid.
        authorized_teams = set(self.authorized_teams)
        query_teams = set()
        now = datetime.datetime.now()
        for team in authorized_teams:
            expiry_date = cache.get(team)
            if expiry_date is None:
                # There's no cache entry for this team, so we'll have to query
                # for it.
                query_teams.add(team)
            elif expiry_date < now:
                # The cache entry for this team has expired.  Evict the cache
                # and query for the team.
                self.debug('Invalidating cache for team: %s', team)
                del cache[team]
                query_teams.add(team)
            else:
                # This team is public.
                self.debug('Cache says, team is public: %s', team)
                pass
        if len(query_teams) == 0:
            # There are no non-public teams to query about.  Short-circuit.
            return True
        # See if there's an XMLRPC server we can use to query about the other
        # teams we couldn't trust our cache about.
        public_team_url = self.options.get('public-team-url')
        if public_team_url is None:
            self.debug('No public team url')
            return False
        public_team_method = self.options.get('public-team-method')
        if public_team_method is None:
            self.debug('No public team method')
            return False
        server = xmlrpclib.ServerProxy(public_team_url)
        method = getattr(server, public_team_method, None)
        if method is None:
            self.error('No method %s() on server: %s',
                       public_team_method, public_team_url)
            return False
        # The cache lifetime is given in seconds, defaulting to 1 hour
        expiration = datetime.datetime.now() + self.cache_lifetime
        # For teams which we couldn't trust our cache about, query the XMLRPC
        # server to see about their visibility.
        for team in query_teams:
            try:
                is_public = method(team)
            except (xmlrpclib.ProtocolError, xmlrpclib.Fault, socket.error):
                sys.excepthook(*sys.exc_info())
                sys.stderr.flush()
                return False
            self.debug('%s is public? %s', team, is_public)
            if not is_public:
                return False
            cache[team] = expiration
        # If we got here, all teams are public.
        return True

    def login_redirect(self, message=None, target=None, logout=False):
        """Issue a 302 redirect to the OpenID login page.

        (mod_python.Request, str or NoneType, str or NoneType) ->
            apache.SERVER_RETURN"""
        if target:
            self.debug('Setting target: %r', target)
            self.session['target'] = target
        self.session['message'] = message
        self.session['logout'] = logout
        self.redirect(self.action_url('login'))

    def autosubmit(self, form, id):
        # Redirect OpenID 2.0 messages by POST'ing a form
        text = post_redirect_tmpl % dict(
            title='OpenID Authentication Required',
            id=id,
            form=form,
            )
        self.apache_request.content_type = 'text/html; charset=UTF-8'
        self.apache_request.set_content_length(len(text))
        self.apache_request.write(text)
        self.session['message'] = None
        return apache.OK

    def redirect(self, url):
        # This function raises an exception, so it will halt anything
        # that calls it. This is probably what you want, but beware!
        #try:
        #    util.redirect(self.apache_request, url)
        #except IOError:
        #    # Mask the symptom of body content being output
        #    # too early. This is not a proper fix, but will
        #    # at least avoid tracebacks.
        #    raise apache.SERVER_RETURN, apache.DONE
 
        # The below will perform a request without printing
        # any default Apache method, but browser cookies
        # will also not be set on the first page load.
        #self.apache_request.log_error("redirecting to: %r" % url)
        self.apache_request.headers_out['location'] = url
        self.apache_request.status = apache.HTTP_MOVED_TEMPORARILY
        raise apache.SERVER_RETURN, apache.DONE

    def action_url(self, action):
        """Generate a URL that performs the given action. This depends
        on knowing where the actions live.
        """
        return urlparse.urljoin(self.server_url, self.action_path + action)


class OpenIDLogin(OpenIDAccessRequest):
    auto_login_lifetime = 24 * 60 * 60 * 365

    def get_consumer(self):
        """Get the consumer that should be used for this request. This is
        also responsible for creating stores. If you want to use a
        different kind of store, change this function to support creating
        it.

        mod_python.Request -> consumer.OpenIDConsumer or apache.SERVER_RETURN
        """
        store_type = self.options.get('store-type', 'memory')
        if store_type == 'file':
            store_dir = self.options['store-directory']
            store_dir = os.path.normpath(store_dir)
            store = self._file_stores.get(store_dir)
            if store is None:
                if len(self._file_stores) > self._max_store_cache_size:
                    self._file_stores.clear()

                store = filestore.FileOpenIDStore(store_dir)
                self._file_stores[store_dir] = store
        elif store_type == 'stateless':
            # stateless dumb store is only available in
            # old versions of the python-openid library
            # (1.0, 1.1 .. not 1.2 or 2.x)
            store = dumbstore.DumbStore("XXX unused")
        elif store_type == 'memory':
            store = memstore.MemoryStore()
        else:
            # XXX: implement other store types (SQL...)
            self.error('Unknown OpenID store type: %r', store_type)
            raise apache.SERVER_RETURN, apache.HTTP_INTERNAL_SERVER_ERROR

        return consumer.Consumer(self.session, store)

    def get_last_user(self):
        cookies = Cookie.get_cookies(self.apache_request)
        cookie = cookies.get('mpopenid.last_user')
        if cookie is None:
            self.debug('No last user')
            return None
        self.debug('Got last user: %r', cookie.value)
        return cookie.value

    def set_last_user(self, username):
        assert isinstance(username, basestring)
        expires = time.time() + self.auto_login_lifetime
        Cookie.add_cookie(self.apache_request,
            'mpopenid.last_user', username, expires=expires)
        self.debug('Set last user: %r', username)

    def del_last_user(self):
        Cookie.add_cookie(
            self.apache_request, 'mpopenid.last_user', '', expires=0)
        self.debug('Deleted last user')

    last_user = property(get_last_user, set_last_user, del_last_user)

    def store_op_for_endpoint(self, op, endpoint):
        """ Store a discovered op=>endpoint mapping for an allowed op """
        store = self.consumer.consumer.store
        assoc = Association.fromExpiresIn(600, op, endpoint, 'HMAC-SHA1')
        if not '://' in op:
            op = 'http://' + op
        store.storeAssociation(op, assoc)

    def fetch_op_for_endpoint(self, endpoint):
        """ Retrieve the (trusted) op that led to a given endpoint, if any """
        store = self.consumer.consumer.store
        allowed_ops = self.get_allowed_ops()
        for op in allowed_ops.values():
            assoc_handle = op
            if not '://' in op:
                assoc_handle = 'http://' + op
            fake_assoc = store.getAssociation(assoc_handle)
            if fake_assoc is None:
                continue
            if fake_assoc.handle == op and fake_assoc.secret == endpoint:
                return op

    def fill_login_page(self, openid_identifier, messages):
        """Generate the HTML for the login page."""
        message_chunks = []
        for name in messages:
            message_info = message_text.get(name)
            if message_info is None:
                message_info = ('error', 'An error occurred')

            chunk = "<div class='%s'>%s</div>" % message_info
            message_chunks.append(chunk)

        if self.cookied_user:
            chunk = ("<div class='message'>You are currently logged "
                     "in as %s. (<a href='%s'>logout</a>)</div>"
                     % (escape(self.cookied_user),
                        escape(self.action_url('logout'), True)))
            message_chunks.append(chunk)

        message_html = '\n'.join(message_chunks)

        target = self.session.get('target')
        if target:
            resource = (
                '<div class="message">Authorization is required to access '
                '<code>%s</code></div>') % (escape(target),)
        else:
            resource = ''

        if not openid_identifier:
            openid_identifier = ''

        allowed_ops = self.get_allowed_ops()
        if len(allowed_ops) == 0:
            # use a text input box
            form_input = '<p>Enter your OpenID identity URL to continue.</p>'
            form_input += ('<input type="text" name="openid_identifier" '
                'value="%s">' % (openid_identifier,))
            onload_js = 'document.oid_form.openid_identifier.focus();'
        elif len(allowed_ops) == 1:
            # use a hidden field
            for k in allowed_ops.keys():
                form_input = ('<input type="hidden" name="openid_identifier" '
                    'value="%s">' % (allowed_ops[k].decode('utf-8'),))
            if target and not self.cookied_user:
               onload_js = ('document.getElementById(\'continue_btn\').style.visibility=\'hidden\'; '
                    + 'document.oid_form.submit();')
            else:
                onload_js = ''
        else:
            # use a dropdown list
            form_input = '<p>Enter your OpenID identity URL to continue.</p>'
            form_input += '<select name="openid_identifier">'
            for k in allowed_ops.keys():
                form_input += ('<option value="%s">'
                    '%s</option>' % (allowed_ops[k], k,))
            form_input += '</select>'
            onload_js = ''

        return login_page_tmpl % dict(
            action=escape(self.action_url('login'), True),
            resource=resource,
            title='OpenID Authentication Required',
            message=message_html,
            openid_identifier=escape(openid_identifier, True),
            form_input=form_input,
            onload=onload_js,
            )

    def get_action_path(self):
        """Find the URL to the actions
        """
        # First check configuration
        action_path = self.options.get('action-path')
        if action_path is None:
            # Default to path for this request (since this handler *is*
            # the action path)
            #
            # Too bad there is no way to just get the Apache path to
            # this handler. It depends on whether this handler appears
            # in a <Files>, <Directory>, or <Location> section. (see
            # <http://mail-archives.apache.org/mod_mbox/httpd-python-dev/200610.mbox/%3C15703531.1161216215467.JavaMail.jira@brutus%3E>)
            path_info = self.apache_request.path_info
            if path_info:
                assert self.apache_request.uri.endswith(path_info)
                action_path = self.apache_request.uri[:-len(path_info)]
            else:
                action_path = self.apache_request.uri

            # XXX: if this is '/' or '' there is probably something
            # wrong with the config because it'd be pretty silly to
            # have this handler at the root of a server. We should
            # tell the user, but how?

        # Make sure that there is a trailing slash on the action path
        if not action_path or action_path[-1] != '/':
            action_path += '/'

        return action_path

    def get_allowed_ops(self):
        """Get configured OP URIs that can be selected
        for login using directed identity.
        """
        op_list = {}
        op_list_url = self.options.get('allowed-op-list-url')
        if op_list_url:
            try:
                url_handle = urllib.urlopen(op_list_url)
            except IOError:
                self.error('Failed to fetch allowed OP URL %r', op_list_url)
                # Write traceback to Apache log
                sys.excepthook(*sys.exc_info())
                sys.stderr.flush()
            else:
                data = url_handle.read()
                url_handle.close()
                op_list = parse_allowed_ops(data)

        return op_list

    def openid(self):
        """Dispatch to the appropriate access control action.
        """
        self.debug('in openid()')

        if self.apache_request.uri.startswith(self.action_path):
            action = self.apache_request.uri[len(self.action_path):]
        # subprocess_env value is too detailed. It includes the
        # query string which is not useful for finding the action.
        #if self.apache_request.subprocess_env['REQUEST_URI'].startswith(self.action_path):
        #    action = self.apache_request.subprocess_env['REQUEST_URI'][len(self.action_path):]
        else:
            self.error(
                'Action path does not match my URL. Configuration problem? '
                '(action_path=%r, uri=%r)',
                self.action_path, self.apache_request.uri)
            self.login_redirect()

        self.debug('openid:do_%s()', action)
        try:
            handler = getattr(self, 'do_' + action)
        except AttributeError:
            # An action we don't know about was called.
            self.error(
                "Unknown OpenID access control action: %r "
                "(action_path=%r)", action, self.action_path)
            self.login_redirect()
        else:
            return handler()

    def do_logout(self):
        try:
            del self.cookied_user
            del self.cookied_teams
            del self.last_user
        except KeyError:
            pass
        self.clear_external_cookies()
        redirect_url = self.options.get('external-logout-url')
        if redirect_url is None:
            self.login_redirect(logout=True)
        else:
            self.redirect(redirect_url)

    def clear_external_cookies(self):
        """Clear any cookies from external apps which are defined in the config
        This is useful when the application protected by Apache OpenID doesn't
        provide a mechanism to log out of OpenID as well as itself.  Redirect
        requests for the application's logout URL should be handled by the
        OpenID logout handler.  The configuration should define the names of
        the application's cookies.  These will be deleted at the same time as
        the OpenID session.
        """
        cookies = Cookie.get_cookies(self.apache_request)
        external_cookies = self.get_external_cookie_names()
        for key in external_cookies:
            if ':' in key:
                path, key = key.split(':', 1)
            else:
                path = '/'
            if key in cookies:
                Cookie.add_cookie(self.apache_request, key, '', expires=0,
                                  path=path)

    def get_external_cookie_names(self):
        """Get all external cookie names for this request."""
        cookie_names_s = self.options.get('external-cookie-names')
        if cookie_names_s is None:
            cookie_names_list = []
            self.debug('No named cookies')
        else:
            cookie_names_list = split_values(cookie_names_s)
            self.debug('Cookie names from config: %s' % cookie_names_list)
        return cookie_names_list

    def do_login(self):
        """Show a login page for setting the OpenID cookie.
        """
        form = util.FieldStorage(self.apache_request)
        openid_identifier = form.getfirst(
            'openid_identifier', self.cookied_user)

        immediate = False
        if self.session.get('logout'):
            del self.session['logout']
        else:
            if not openid_identifier and self.last_user:
                openid_identifier = self.last_user
                # immediate mode messes up redirects?
                #immediate = True

        message = self.session.get('message', None)
        if message is None:
            messages = []
        else:
            messages = [message]

        # Check that the requested identity is allowed
        allowed_ops = self.get_allowed_ops()
        if not openid_identifier in allowed_ops.values():
            openid_identifier = ''
        if self.apache_request.method == 'POST' or immediate:
            if openid_identifier:
                try:
                    auth_request = self.consumer.begin(openid_identifier)
                except expected_discovery_exceptions:
                    # Print traceback to Apache log (stderr goes there)
                    sys.excepthook(*sys.exc_info())
                    sys.stderr.flush()

                    # Note that this was an exception in discovery
                    messages.append('discovery')
                else:
                    # Store the discovered endpoint
                    self.store_op_for_endpoint(openid_identifier,
                                  auth_request.endpoint.server_url)
                    # Request Launchpad teams information, if configured
                    if teams_extension_avail \
                            and len(self.authorized_teams) > 0:
                        teams_request = teams.TeamsRequest(
                            self.authorized_teams)
                        auth_request.addExtension(teams_request)

                    if auth_request.shouldSendRedirect():
                        # Do the redirect to the OpenID server
                        self.debug("Redirecting to OP with GET")
                        redirect_url = auth_request.redirectURL(
                            self.server_url,
                            self.action_url('return'),
                            immediate)
                        self.redirect(redirect_url)
                    else:
                        # use a form-based redirect
                        self.debug("Redirecting to OP with POST")
                        form_html = auth_request.formMarkup(
                            self.server_url,
                            self.action_url('return'),
                            form_tag_attrs={'id':'openid_message'},
                            immediate=immediate)
                        self.autosubmit(form_html, 'openid_message')

                    return apache.OK
            else:
                messages.append('empty')

        text = self.fill_login_page(openid_identifier, messages)
        self.apache_request.content_type = 'text/html; charset=UTF-8'
        self.apache_request.set_content_length(len(text))
        self.apache_request.write(text)
        self.session['message'] = None
        return apache.OK

    def do_return(self):
        """Handle a response from the OpenID server. Always redirects.
        mod_python.Request -> apache.SERVER_RETURN
        """
        form = util.FieldStorage(self.apache_request)
        query = {}
        for k in form.keys():
            query[k] = form.getfirst(k).decode('utf-8')

        response = self.consumer.complete(query, self.action_url('return'))

        if response is None:
            self.login_redirect(message='failure')
        elif response.status == consumer.SUCCESS:
            # Set the cookie and then redirect back to the target
            self.last_user = self.cookied_user = response.identity_url

            # Check if the endpoint is one we trust
            op = self.fetch_op_for_endpoint(response.endpoint.server_url)
            if op is None:
                self.login_redirect('failure')
            else:
                # Check for Launchpad teams data in response
                self.debug('do_return: teams_extension_avail: %s',
                           teams_extension_avail)
                self.debug('do_return: authorized_teams: %r',
                           self.authorized_teams)
                if teams_extension_avail and len(self.authorized_teams) > 0:
                    teams_response = teams.TeamsResponse.fromSuccessResponse(
                        response)
                    self.debug('teams_response.is_member = %r',
                               teams_response.is_member)
                    self.cookied_teams = teams_response.is_member
                self.debug('do_return: cookied_teams: %r', self.cookied_teams)

                target = self.session.get('target')
                if not target:
                    self.debug('session target not set. '
                               'redirecting to login page: %r',
                               self.action_url('login'))
                    target = self.action_url('login')
                else:
                    self.debug('redirecting to session target %r', target)
                self.session['target'] = None
                self.redirect(target)
        elif response.status == consumer.CANCEL:
            self.login_redirect(message='cancel')
        elif response.status == consumer.FAILURE:
            self.login_redirect(message='failure')
        elif response.status == consumer.SETUP_NEEDED:
            # Silently redirect to login
            del self.last_user
            self.login_redirect()
        else:
            assert False, response.status

class OpenIDProtect(OpenIDAccessRequest):
    def get_authorized_urls(self):
        """Get all authorized URLs for this request
        """
        authorized_s = self.options.get('authorized-users', None)
        if authorized_s is None:
            authorized_list = []
        else:
            try:
                authorized_list = self._authorized_cache[authorized_s]
            except KeyError:
                if (len(self._authorized_cache) ==
                    self._max_authorized_cache_size):
                    self._authorized_cache.clear()

                authorized_list = parse_authorized_urls(authorized_s)
                self._authorized_cache[authorized_s] = authorized_list

        authorized_list_url = self.options.get(
            'authorized-users-list-url')
        if authorized_list_url is not None:
            try:
                url_handle = urllib.urlopen(authorized_list_url)
            except IOError:
                self.error('Failed to fetch authorized list URL %r',
                           authorized_list_url)
                # Write traceback to Apache log
                sys.excepthook(*sys.exc_info())
                sys.stderr.flush()
            else:
                data = url_handle.read()
                url_handle.close()
                authorized_list.extend(parse_authorized_urls(data))

        return authorized_list

    def get_action_path(self):
        """Find the URL to the actions
        """
        action_path = self.options.get('action-path')
        if action_path is None:
            # the path where *Handler directive was specified
            protected_path = self.apache_request.hlist.directory
            if protected_path:
                docroot = self.apache_request.document_root()
                protected_path = protected_path[len(docroot):]
                if not protected_path or protected_path[-1] != '/':
                    protected_path += '/'
            else:
                self.error(
                    'No action-path specified and the protect directive is '
                    'in a <Location> block. Defaulting to /openid for login '
                    'actions.')
                protected_path = '/'

            action_path = protected_path + 'openid/'
        elif not action_path or action_path[-1] != '/':
            action_path += '/'

        return action_path

    def protect(self):
        """Only allow the request to proceed if the openid_identifier
        has authenticated with OpenID and is in the authorized list.
        The list can be in the authorized-users setting, with a
        supplemental list specified by the authorized-users-list-url,
        which should be text/plain and be a list of OpenIDs, one per
        line.

        mod_python.Request -> int or apache.SERVER_RETURN
        """
        self.debug('in protect()')

        # Do not apply rule if this is handled by one of the OpenID
        # access control actions.
        # Also ignore special case of "favicon.ico" which browsers
        # request automatically, and will incorrectly update the
        # session target value.
        #if self.apache_request.uri.startswith(self.action_path):
        env = self.apache_request.subprocess_env
        if (env['REQUEST_URI'].startswith(self.action_path)
            or self.request_uri.endswith('favicon.ico')):
            self.debug('Protection requested for OpenID endpoint or favicon, '
                       'returning apache.OK')
            return apache.OK

        # Allow public teams with no further login required.
        if self.check_public_teams():
            self.debug('Public teams allowed by all: %r',
                       self.authorized_teams)
            return apache.OK

        # Check to see if cookied openid_identifier is authorized
        if self.cookied_user:
            if self.cookied_user in self.authorized_urls:
                self.apache_request.user = self.cookied_user

                # Do not cache this access-controlled page
                self.apache_request.headers_out['Cache-Control'] = 'no-cache'
                self.debug('Allowed user %r', self.cookied_user)
                return apache.OK
            # Check to see if any of the cookied Launchpad teams are
            # authorized.
            if teams_extension_avail and len(self.authorized_teams) > 0:
                self.debug("Authorized Teams: %r",
                           ', '.join(self.authorized_teams))
                self.debug("Cookied Teams: %r",
                           ', '.join(self.cookied_teams))
                if self.cookied_teams and len(self.cookied_teams) > 0:
                    for team in self.cookied_teams:
                        if team in self.authorized_teams:
                            self.apache_request.user = self.cookied_user
                            self.apache_request.headers_out[
                                'Cache-Control'] = 'no-cache'
                            self.debug('Allowed user %r - team %r',
                                       self.cookied_user, team)
                            return apache.OK
            else:
                self.debug('Team support disabled, or no authorized teams '
                           'configured.')
            self.error(
                'Unauthorized access attempt for %r by %r (teams: %r)',
                self.request_uri, self.cookied_user, self.cookied_teams)
            message = 'denied'
        else:
            # Initial request with no openid_identifier cookie, so no message.
            message = None

        # The redirects only work for GET, so just return FORBIDDEN if
        # it's any other method.
        if self.apache_request.method != 'GET':
            raise apache.SERVER_RETURN, apache.HTTP_FORBIDDEN

        # cookied_user not authorized or not set, so redirect to login
        # and storing the original uri requsted to use after auth
        self.login_redirect(message, self.request_uri)
